#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <set>
#include <map>
#include <iomanip>
#include <limits>
#include <unordered_map>
#include <algorithm>
#include <cctype>
#include "class_header.hpp"
#include "amino_nucleic_acids.hpp"
#include "SSalgorithm.hpp"


int main (int argc, char** argv){
	ss::citation();

 if (argc < 2){
	 std::cerr <<"To use HIT-2 should at lease have a inputfile" << std::endl;
 exit(1);
 }

    std::string filename = std::string(argv[1]); //change the filename as cpp string
    std::string filebasename; 


/////////////////////////////////////////////////////
/////remove the extension of filename////////////////
    size_t dot_pos = filename.find(".");
if (dot_pos != std::string::npos) {
    filebasename = filename.substr(0, dot_pos);
//    std::cout<<filebasename<<std::endl;
}else{ filebasename = filename; 
    std::cout<<"no extention, the filebasename is: "<<filename<<std::endl;}
    std::string biostrname = "HIT-2-" + filebasename +"-biomolecule"+".pdb"; 
    std::ofstream HIT2bio(biostrname); 
    std::string comstrname = "HIT-2-" + filebasename +"-complex"+".pdb"; 
    std::ofstream HIT2com(comstrname); 
    // Create file
/////remove the extension of filename////////////////  
////////////////////////////////////////////////////


std::ifstream pdb_file(filename);
  if (!pdb_file) {
        std::cerr << "Error: Unable to open file: " << filename << "\n";
        exit(1);
    }
       
/////////////////////////////////////////////////////////////////////////////////////////////// 
//////////////File checking///////////////////////////////////////////////////////////////////   
////checking number of frames./////////////////////////
 int num_model = 1;  // initialize the model counter
 int num_line = 1; // initialize the num_line counter
std::string first_line;
 while (std::getline(pdb_file, first_line)) {
     if (first_line.substr(0, 4) == "ATOM" || first_line.substr(0, 6) == "HETATM") {
//         std::cout<<"The ref line is:" << first_line.substr(0,22) << std::endl;
         break;
     }
 }
/////get the first line as the indicator for new frame identifier

////////checking number of frames and atoms.//////////////////
 std::string line;
 while (std::getline(pdb_file, line)) {
    if (line.substr(0, 6) == "ATOM  " || line.substr(0, 6) == "HETATM") {
            ++num_line;
     //check first 22 letters(until chainID to make sure the new frame start ),
        if (first_line.substr(0, 22) == line.substr(0, 22)) {
            ++num_model;
        }
     }
 }  // checking if the serial number is renamed
 std::cout <<"Number of frames: " <<num_model << std::endl;
//////checking number of frames.//////////////////


///checking if file is complete/////
 std::cout <<"Number of atom lines: " <<num_line << std::endl;
 std::cout <<"Number of atom in each frame: " <<std::setprecision(10)<<num_line/(double)num_model << std::endl;
if (num_line%num_model!=0){
	std::cout<<RED<<"Serious Warning: The file maybe incomplete, Different number of ATOMs in frames"<<std::endl;
        exit(1);
       }

int num_perframe = num_line/num_model;
std::cout <<"Number of atom in each frame is set well: " << num_perframe << std::endl;
std::cout<< MAGENTA <<"Good News: Same number of atoms in each frame"<<RESET<<std::endl;
//////////////File checking////////////////////////////////////////////////////////////////////   
/////////////////////////////////////////////////////////////////////////////////////////////// 


 
int target_frameID = num_model;
double iterationstep = 0.2;
double sigma = 0.95;
  if( argc == 3  ){
 target_frameID = atoi(argv[2]);
 }else if (argc == 4 ) {
  target_frameID = atoi(argv[2]);
  iterationstep  = atof(argv[3]);
 }else if (argc == 5 ) {
  target_frameID = atoi(argv[2]);
  iterationstep  = atof(argv[3]);
  sigma          = atof(argv[4]);
 }else if (argc >=6){std::cerr << "S.S Warning: Too many arguments, my lord"<< std::endl;}


///////////////////////////////////////////////////////////////////////////////////////////
/// parameters checking.///////////////  
 if( target_frameID <= 0 ){std::cout<<BLUE<<"S.S knows you. You want the first frame, right?"<<std::endl;
                          target_frameID=1;             }
 if( target_frameID > num_model){std::cout<<BLUE<<"S.S knows you. You want the last frame, right?"<<std::endl;
                          target_frameID=num_model;  }
 if( iterationstep < 0 ){std::cout<<RED<<"S.S warning. WTF are you doing? I have reset the step size as 0.2!" << std::endl;
                          iterationstep=0.2;  }
 if( iterationstep >= 1  ){std::cout<<RED<<"S.S warning. Be careful, your iterationstep is too big" << std::endl; }
 if(sigma <= 0) { 
	 std::cout<<RED<<"S.S warning. WTF are you doing? I have reset the filling ratio threshold as 0.95!" << std::endl;
                          sigma = 0.95;     }
 
 else if(  sigma > 1 || sigma < 0.5  ){std::cout<<MAGENTA <<"S.S warning. Be careful, the Filling ratio threshold should be belonging to [0.5,1], but I respect your choice, my lord~ " << std::endl; }
///checking parameters.///////////////

///Output parameters setting/////////////////
	std::cout <<RESET<< "Input file is: " << argv[1]<<std::endl;
        std::cout << "Target biomolecular structure's frame ID is: "<< target_frameID << std::endl;
        std::cout << "Step size is: " << iterationstep << std::endl;
        std::cout << "Filling ratio threshold is: " << sigma << std::endl;
///Output parameters setting/////////////////
////////////////////////////////////////////////////////////////////////////////////////////// 

 
///re-read files///////////////////////////////////////
pdb_file.clear();
pdb_file.seekg(0,std::ios::beg);
///re-read files///////////////////////////////////////


std::cout<<CYAN<<" PDB file reading start "<<RESET<<std::endl;


///////////////////////////////////////////////////////////////
//////Store information into ATOM Class////////////////////////
num_model=0;//initialize to 0 for better searching
std::vector<ATOM> atoms;//pdb file reading
std::vector<point3D> masscenter;//str masscenters of all frames
std::set<std::string> undefined_aminoacid;
while (std::getline(pdb_file, line)) {
    if (line.substr(0, 6) == "ATOM  " || line.substr(0, 6) == "HETATM") {
         int index = std::stoi(line.substr(6, 5));
         std::string name = line.substr(12, 4);
          name.erase(std::remove_if(name.begin(),name.end(),::isspace),name.end());
         std::string aminoAcid = line.substr(17, 3);
          aminoAcid.erase(std::remove_if(aminoAcid.begin(),aminoAcid.end(),::isspace),aminoAcid.end());
         std::string chainId = line.substr(21, 1);
         int resindex = std::stoi(line.substr(22,4)); 
         double x = std::stod(line.substr(30, 8));
         double y = std::stod(line.substr(38, 8));
         double z = std::stod(line.substr(46, 8));
         double alpha = std::stod(line.substr(54, 6));
         double beta = std::stod(line.substr(60, 6));
         std::string specifier;

         //define a class atom;
         ATOM atom(index, name, aminoAcid, chainId,resindex, x, y, z, alpha, beta,specifier);
 

    if (aminoAcids.count(atom.aminoAcid)) {
        atom.specifier = "A";
    } else if (nucleicAcids.count(atom.aminoAcid)) {
        atom.specifier = "N";
    } else if (Cation.count(atom.aminoAcid)) {
        atom.specifier = "CAT";
    } else if (Anion.count(atom.aminoAcid)) {
        atom.specifier = "ANI";
    } else {
       undefined_aminoacid.insert(atom.aminoAcid);
    }
            // Add the atom to the vector
            atoms.push_back(atom);
	    if (first_line.substr(0, 22) == line.substr(0, 22)) {
            ++num_model;
//         std::cout <<"Number of frames: " <<num_model << std::endl;
        }
     }//reading if ATOM
 }//reading done

std::cout<<CYAN<<" PDB file read successfully! "<<RESET<<std::endl;
/////Store information into ATOM Class/////////////////////////
///////////////////////////////////////////////////////////////
if (undefined_aminoacid.empty()){
	std::cout<<GREEN<<"All ions or amino acids have been identified"<<RESET<<std::endl;
}else{
    std::cout<<RED<<"S.S. Warning: Undefined residue or ions found: "<<RESET<<std::endl;
	 for (const auto&str: undefined_aminoacid){   
		 std::cout   <<str<<std::endl;
	 }
	 std::cout<<  "Add them to \"amino_nucleic_acids.h\" and compile it by g++ HIT2cpp.cpp -o HIT2cpp.o, if you want to calculate these ion" << std::endl;
}
////////////////////////////////////////////////////////////////
///////////////target structure stored//////////////////////////
for (int i=0; i<num_perframe; i++){
    int index =num_perframe*(target_frameID-1) + i;
    if (index < atoms.size()) {
	    if(atoms[index].specifier!="CAT" &&  atoms[index].specifier!="ANI" ){
        atoms.at(index).output2file(HIT2bio);
        atoms.at(index).output2file(HIT2com);
//        std::cout << "Biostrfile written successfully."<<std::endl;
        }
    } else {
        std::cerr << "Index out of range: " << index << std::endl;
    }
}


if (HIT2bio.is_open()) { // Check if file was opened successfully
        HIT2bio.close(); // Close file
    }
////////////////////////////////////////////////////////////////
///////////////target structure stored//////////////////////////


///////////////////////////////////////////////////////////////
//////////////Masscenter information///////////////////////////
//

//std::ofstream mass("mass.txt");
for (int i = 0; i< num_model; i++){
 double tx=0, ty=0, tz=0;
 double num_stratom = 1.0; //In case there is no protein atom
 for (int j = 0; j< num_perframe; j++ ){
	   int index=i*num_perframe+j;
  if(atoms[index].specifier !="CAT" &&  atoms[index].specifier !="ANI" ){
   tx+=atoms[index].x;
   ty+=atoms[index].y;
   tz+=atoms[index].z;
   ++num_stratom;    
  }
 }
 double mx = tx/num_stratom;
 double my = ty/num_stratom;
 double mz = tz/num_stratom;
 point3D mc(mx,my,mz);
// mc.output2file(mass);
 masscenter.push_back(mc); 
}

//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////

/// check how many type of ions /////////////////////////////
/////////////////////////////////////////////////////////////

std::set<std::string> unique_ion;
for (int i = 0; i < num_perframe;i++ ) {
   if (atoms[i].specifier=="CAT" || atoms[i].specifier == "ANI")
    unique_ion.insert(atoms[i].aminoAcid);
}
// Print out the unique names
//
    if (unique_ion.empty()) {
   std::cout <<RED<< "No ions in your pdb file" <<RESET<<std::endl;
	exit(1);
   }else{
   std::cout <<RED<< "The ion types include: "<<RESET<<std::endl;
       for (const auto& str: unique_ion){
         	std::cout<<str<<std::endl;
}}
/////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////
//////////////Store ions posistion////////////////////////////
std::map<std::string, std::vector<ion3D>> ionpos;
for (const auto& name : unique_ion){
//	std::ofstream ionpos_file (name+".postxt");
 std::vector<ion3D> tempVec;
 ionpos[name] = tempVec;
// ionpos_file.close();
}

for (int i = 0; i< num_model; i++){
  for (int j = 0; j< num_perframe; j++ ){
  int index=i*num_perframe+j;
  auto it = unique_ion.find(atoms[index].aminoAcid);
	  if(it != unique_ion.end() ){
                double ix=atoms[index].x-masscenter[i].x;
                double iy=atoms[index].y-masscenter[i].y;
                double iz=atoms[index].z-masscenter[i].z;
		std::string iontype =  *it;
		ion3D Newion(ix,iy,iz);
                ionpos[iontype].push_back(Newion);
	  }

   }
}


//for (const auto& name : unique_ion){
// std::ofstream ionpos_file (name+".postxt");
// 
//   if (!ionpos_file.is_open()) {
//    std::cerr << "Failed to open output file!" << std::endl;
//    return 1;
//  }
//  
//  // loop over the vector and call output2file for each element
//  for (auto& ion : ionpos[name]) {
//    ion.output2file(ionpos_file);
//  } 
// 
// ionpos_file.close();
//}


//////////clear memory////////////////////////////////////////
atoms.clear();
//////////////////////////////////////////////////////////////



///////////////   Screening  /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////



///// Checking the Min and Max of the Each ions //////////////
///// And Setting the a map of DensityMap       //////////////
///// And Setting the a map of BoundIons        //////////////
std::map<std::string, MinMax> xyz_limits;
std::map<std::string, std::vector<ATOM>> BoundIons;

for (const auto& pair : ionpos) {
        const std::vector<ion3D>& ions = pair.second;
	// create BoundIons map same time;
	std::vector<ATOM>& Bound = BoundIons[pair.first];
	// create an XYZ object for this vector
	MinMax& xyz = xyz_limits[pair.first];
        // loop through each ion3D in the vector
        for (const auto& ion : ions) {
            xyz.update(ion.x, ion.y, ion.z);
        }
    }
//for (const auto& pair: xyz_limits) {
//	std::cout<<pair.first<<std::endl;
//	pair.second.print();
//    }
///// Checking the Min and Max of the Each ions //////////////
///// And Setting the a map of DensityMap       //////////////
///// And Setting the a map of BoundIons        //////////////


//////////  Iterations   //////////////////////////////////////

for (auto&iontype : unique_ion){
  std::cout<<YELLOW<<iontype<<" Calculation Start " <<RESET<< std::endl;
  for (double cs=INITIAL_CUBE_SIZE; cs < END_CUBE_SIZE ; cs+=iterationstep){
  std::cout<<"cs is "<<cs<<std::endl;
  std::vector<DensityMap>DenM=ss::DenMB(xyz_limits[iontype],cs);
  ss::trans(DenM, ionpos[iontype], xyz_limits[iontype],cs);
  // for (auto &denm: DenM.begin()->second){
  //	 std::cout << "the n is " << denm.n<<std::endl; }
  
  ss::SaReIons(iontype,"I",DenM, ionpos[iontype], BoundIons[iontype], xyz_limits[iontype], masscenter[target_frameID-1],cs,(int)(sigma*num_model) );
  std::vector<DensityMap>().swap(DenM);
  
  }
  //////////  Iterations   //////////////////////////////////////
  
    std::string ionfile="HIT-2_bound_"+iontype+".pdb";
    std::ofstream IonFile(ionfile);
    for ( auto&sion: BoundIons[iontype]){
    sion.output2file(IonFile);
    sion.output2file(HIT2com);
    }
    IonFile.close();
   std::cout<<iontype<<" Calculation END " << std::endl;
}
    


///////////////   END  ////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
ionpos.clear();
masscenter.clear();

ss::citation();

return 0;

}






