#ifndef ATOM_H
#define ATOM_H


class ATOM {
public:
    int index;
    std::string name;
    std::string aminoAcid;
    std::string chainId;
    int resindex;
    double x;
    double y;
    double z;
    double alpha;
    double beta;
    std::string specifier;
    // Constructor
    ATOM(int i,
         std::string n,
         std::string aa,
         std::string chain,
	 int j,
         double x,
         double y,
         double z,
         double a,
         double b,
	 std::string sp) :
         index(i), name(n), aminoAcid(aa),  chainId(chain), resindex(j),x(x), y(y), z(z), alpha(a), beta(b),specifier(sp) {}
         
    // output parameters
    void output() {
        std::cout << "Index: "           << index         << std::endl;
        std::cout << "Atom name: "       << name          << std::endl;
        std::cout << "Amino acid name: " << aminoAcid     << std::endl;
        std::cout << "x-coordinate: "    << x             << std::endl;
        std::cout << "y-coordinate: "    << y             << std::endl;
        std::cout << "z-coordinate: "    << z             << std::endl;
        std::cout << "Chain ID: "        << chainId       << std::endl;
        std::cout << "Alpha: "           << alpha         << std::endl;
        std::cout << "Beta: "            << beta          << std::endl;
        std::cout << "Specifier: "       << specifier     << std::endl;
    
    }

   void outputPDB(){
           std::cout << std::left << std::setw(6) << "ATOM"
                  << std::right << std::setw(5) << index
                  << " " << std::setw(4) << name
                  << " " << std::setw(3) << aminoAcid
                  << " " << std::setw(1) << chainId
                  << std::right << std::setw(4) << resindex
                  << "    "
                  << std::fixed << std::setprecision(3)
                  << std::setw(8) << x
                  << std::setw(8) << y
                  << std::setw(8) << z
                  << std::setw(6) << std:: setprecision(2) << alpha 
                  << std::setw(6) << std:: setprecision(2) << beta
                  << std::endl;
  
   }
   void output2file(std::ofstream& outfile){
          outfile << std::left << std::setw(6) << "ATOM"
                  << std::right << std::setw(5) << index
                  << " " << std::setw(4) << name
                  << " " << std::setw(3) << aminoAcid
                  << " " << std::setw(1) << chainId
                  << std::right << std::setw(4) << resindex
                  << "    "
                  << std::fixed << std::setprecision(3)
                  << std::setw(8) << x
                  << std::setw(8) << y
                  << std::setw(8) << z
                  << std::setw(6) << std:: setprecision(2) << alpha 
                  << std::setw(6) << std:: setprecision(2) << beta
                  << " " << specifier << std::endl;
  
   }


    // Destructor
    ~ATOM(){}
};

class ion3D {
	public:
	 double x, y, z;
	 int ix, iy, iz;
      ion3D(
         double x,
         double y,
         double z,
	 int ix=-1,
	 int iy=-1,
	 int iz=-1):
         x(x),y(y),z(z),ix(ix),iy(iy),iz(iz){}

       void CalIndex(double minx,double miny, double minz, double cubesize){
       this-> ix=(this->x-minx)/cubesize; 
       this-> iy=(this->y-miny)/cubesize; 
       this-> iz=(this->z-minz)/cubesize; 
       }
       
       void refresh(){
        this-> ix=-1; 
        this-> iy=-1; 
        this-> iz=-1; 
       }
       void output2file(std::ofstream& outfile){
        outfile  << std::setw(6)<<std::setprecision(3)<<std::fixed
		  << " "    << x
                  << " "    << y
                  << " "    << z 
		  << " "    << ix
		  << " "    << iy
		  << " "    << iz << std::endl;
   }
      void output(){
        std::cout << std::setw(6)<<std::setprecision(3)<<std::fixed
		  << " "    << x
                  << " "    << y
                  << " "    << z 
		  << " "    << ix
		  << " "    << iy
		  << " "    << iz << std::endl;
   }



};

class point3D {
	 public:
	 double x, y, z;
      point3D(
         double x,
         double y,
         double z):
         x(x),y(y),z(z){}
   void output() {
        std::cout << std::setw(6)<<std::setprecision(2)<<std::fixed
		  << " "    << x
                  << " "    << y
                  << " "    << z << std::endl;
    }
   void output2file(std::ofstream& outfile){
        outfile  << std::setw(6)<<std::setprecision(2)<<std::fixed
		  << " "    << x
                  << " "    << y
                  << " "    << z << std::endl;
   }
};

class DensityMap {
	public: 
        int ix, iy, iz, n;
        DensityMap(
	int x,
	int y,
	int z,
	int n):
	ix(x),iy(y),iz(z),n(n) {}	

   void output() {
        std::cout << std::setw(6)<<std::setprecision(2)<<std::fixed
		  << " "    << ix
                  << " "    << iy
                  << " "    << iz  
		  << " "    << n << std::endl;
    }
   void output2file(std::ofstream& outfile){
        outfile  << std::setw(6)<<std::setprecision(2)<<std::fixed
		  << " "    << ix
                  << " "    << iy
                  << " "    << iz 
		  << " "    << n << std::endl;
   }


};


class MinMax {
public:
    double x_min = std::numeric_limits<double>::max();
    double y_min = std::numeric_limits<double>::max();
    double z_min = std::numeric_limits<double>::max();
    double x_max = std::numeric_limits<double>::lowest();
    double y_max = std::numeric_limits<double>::lowest();
    double z_max = std::numeric_limits<double>::lowest();

    void update(double x, double y, double z) {
        if (x < x_min) x_min = x;
        if (y < y_min) y_min = y;
        if (z < z_min) z_min = z;
        if (x > x_max) x_max = x;
        if (y > y_max) y_max = y;
        if (z > z_max) z_max = z;
    }

    void print() const {
        std::cout << "x: " << x_min << " " << x_max << std::endl << 
		     "y: " << y_min << " " << y_max << std::endl <<
		     "z: " << z_min << " " << z_max << std::endl;
    }

};





#endif // ATOM_H

